@@ -265,7 +265,7 @@ class OutlierTree:
         Parameters
         ----------
         df : DataFrame(n_rows, n_cols)
-            Pandas' DataFrame with normal data that might contain some outliers.
+            Pandas' DataFrame with regular (i.e. non-outlier) data that might contain some outliers.
         cols_ignore : boolean array(n_cols, ) or string array(n_ignore, )
             Array containing columns which will not be split, but will be evaluated for usage in splitting other columns.
             Can pass either a boolean array with the same number of columns as 'df', or a list/array of column names
@@ -536,6 +536,7 @@ class OutlierTree:
         ### Booleans are taken as categoricals
         if df_bool.shape[1] > 0:
             self.cols_bool_ = df_bool.columns.values.copy()
+            self._cat_mapping += [np.array([False,True])]*df_bool.shape[1]
             if df_cat is not None:
                 df_cat = pd.concat([df_cat, df_bool], axis = 1)
             else:
